@@ -1,11 +1,11 @@
 import sys
 import trio
+import types
 import asyncio
-import sniffio
-try:
-    from asyncio.format_helpers import _format_callback, _get_function_source
-except ImportError:  # <3.7
-    from asyncio.events import _format_callback, _get_function_source
+from asyncio.format_helpers import _format_callback, _get_function_source
+
+if sys.version_info < (3, 11):
+    from exceptiongroup import BaseExceptionGroup
 
 
 def _format_callback_source(func, args):
@@ -54,6 +54,65 @@ class ScopedHandle(asyncio.Handle):
         self._loop.call_exception_handler(context)
 
 
+# copied from trio._core._multierror, but relying on traceback constructability
+# from Python (as introduced in 3.7) instead of ctypes hackery
+def concat_tb(head, tail):
+    # We have to use an iterative algorithm here, because in the worst case
+    # this might be a RecursionError stack that is by definition too deep to
+    # process by recursion!
+    head_tbs = []
+    pointer = head
+    while pointer is not None:
+        head_tbs.append(pointer)
+        pointer = pointer.tb_next
+    current_head = tail
+    for head_tb in reversed(head_tbs):
+        current_head = types.TracebackType(
+            current_head, head_tb.tb_frame, head_tb.tb_lasti, head_tb.tb_lineno
+        )
+    return current_head
+
+
+# copied from trio._core._run with minor modifications:
+def collapse_exception_group(excgroup):
+    """Recursively collapse any single-exception groups into that single contained
+    exception.
+    """
+    exceptions = list(excgroup.exceptions)
+    modified = False
+    for i, exc in enumerate(exceptions):
+        if isinstance(exc, BaseExceptionGroup):
+            new_exc = collapse_exception_group(exc)
+            if new_exc is not exc:
+                modified = True
+                exceptions[i] = new_exc
+
+    if len(exceptions) == 1 and getattr(excgroup, "collapse", False):
+        exceptions[0].__traceback__ = concat_tb(
+            excgroup.__traceback__, exceptions[0].__traceback__
+        )
+        return exceptions[0]
+    elif modified:
+        return excgroup.derive(exceptions)
+    else:
+        return excgroup
+
+
+def collapse_aware_exception_split(exc, etype):
+    if not isinstance(exc, BaseExceptionGroup):
+        if isinstance(exc, etype):
+            return exc, None
+        else:
+            return None, exc
+
+    match, rest = exc.split(etype)
+    if isinstance(match, BaseExceptionGroup):
+        match = collapse_exception_group(match)
+    if isinstance(rest, BaseExceptionGroup):
+        rest = collapse_exception_group(rest)
+    return match, rest
+
+
 class AsyncHandle(ScopedHandle):
     """A ScopedHandle associated with the execution of a Trio-flavored
     async function.
@@ -79,30 +138,10 @@ class AsyncHandle(ScopedHandle):
                     self.cancel()
 
     async def _run(self):
-        sniffio.current_async_library_cvar.set("trio")
         self._started.set()
         if self._cancelled:
             return
 
-        def report_exception(exc):
-            if not isinstance(exc, Exception):
-                # Let BaseExceptions such as Cancelled escape without being noted.
-                return exc
-            # Otherwise defer to the asyncio exception handler. (In an async loop
-            # this will still raise the exception out of the loop, terminating it.)
-            self._raise(exc)
-            return None
-
-        def remove_cancelled(exc):
-            if isinstance(exc, trio.Cancelled):
-                return None
-            return exc
-
-        def only_cancelled(exc):
-            if isinstance(exc, trio.Cancelled):
-                return exc
-            return None
-
         try:
             # Run the callback
             with self._scope:
@@ -117,19 +156,24 @@ class AsyncHandle(ScopedHandle):
 
         except BaseException as exc:
             if not self._fut:
-                # Pass Exceptions through the fallback exception handler since
-                # they have nowhere better to go. Let BaseExceptions escape so
-                # that Cancelled and SystemExit work reasonably.
-                with trio.MultiError.catch(report_exception):
-                    raise
+                # Pass Exceptions through the fallback exception
+                # handler since they have nowhere better to go. (In an
+                # async loop this will still raise the exception out
+                # of the loop, terminating it.) Let BaseExceptions
+                # escape so that Cancelled and SystemExit work
+                # reasonably.
+                rest, base = collapse_aware_exception_split(exc, Exception)
+                if rest:
+                    self._raise(rest)
+                if base:
+                    raise base
             else:
                 # The result future gets all the non-Cancelled
                 # exceptions.  Any Cancelled need to keep propagating
                 # out of this stack frame in order to reach the cancel
-                # scope for which they're intended.  This would be a
-                # great place for ExceptionGroup.split() if we had it.
-                cancelled = trio.MultiError.filter(only_cancelled, exc)
-                rest = trio.MultiError.filter(remove_cancelled, exc)
+                # scope for which they're intended.  Any non-Cancelled
+                # BaseExceptions keep propagating.
+                cancelled, rest = collapse_aware_exception_split(exc, trio.Cancelled)
                 if not self._fut.cancelled():
                     if rest:
                         self._fut.set_exception(rest)
@@ -137,6 +181,7 @@ class AsyncHandle(ScopedHandle):
                         self._fut.cancel()
                 if cancelled:
                     raise cancelled
+
         finally:
             # asyncio says this is needed to break cycles when an exception occurs.
             # I'm not so sure, but it doesn't seem to do any harm.
