@@ -1,18 +1,16 @@
 # This code implements a clone of the asyncio mainloop which hooks into
 # Trio.
 
+import os
 import sys
 import trio
 import asyncio
 import threading
 from contextvars import ContextVar
-
-try:
-    from contextlib import asynccontextmanager
-except ImportError:
-    from async_generator import asynccontextmanager
+from contextlib import asynccontextmanager
 
 from ._async import TrioEventLoop
+from ._util import run_aio_future
 from ._deprecate import warn_deprecated
 
 try:
@@ -185,6 +183,25 @@ class _TrioPolicy(asyncio.events.BaseDefaultEventLoopPolicy):
             super().set_event_loop(loop)
 
 
+# get_event_loop() without a running loop is deprecated in 3.12+. The logic for emitting the
+# DeprecationWarning walks the stack looking at module names in order to associate it with
+# the first caller outside asyncio. We need to pretend to be asyncio in order for that to work.
+if sys.version_info >= (3, 12):
+    __name__ = "asyncio.fake.trio_asyncio._loop"
+
+# Make sure we don't try to continue using the Trio loop after a fork()
+if hasattr(os, "register_at_fork"):
+
+    def _clear_state_after_fork():
+        if _in_trio_context():
+            from trio._core._run import GLOBAL_RUN_CONTEXT
+
+            del GLOBAL_RUN_CONTEXT.task
+            del GLOBAL_RUN_CONTEXT.runner
+            current_loop.set(None)
+
+    os.register_at_fork(after_in_child=_clear_state_after_fork)
+
 from asyncio import events as _aio_event
 
 #####
@@ -204,9 +221,14 @@ def _new_policy_set(new_policy):
         raise RuntimeError("You can't set the Trio loop policy manually")
     if _in_trio_context():
         raise RuntimeError("You can't change the event loop policy in Trio context")
-    else:
-        assert new_policy is None or isinstance(new_policy, asyncio.AbstractEventLoopPolicy)
-        _faked_policy.policy = new_policy
+    if (new_policy is not None and not isinstance(new_policy, asyncio.AbstractEventLoopPolicy)):
+        # Raise the type of error that the CPython test suite expects
+        raise_type = TypeError if sys.version_info >= (3, 11) else AssertionError
+        raise raise_type(
+            "policy must be an instance of AbstractEventLoopPolicy or None, "
+            f"not '{type(new_policy).__name__}'"
+        )
+    _faked_policy.policy = new_policy
 
 
 _orig_policy_get = _aio_event.get_event_loop_policy
@@ -218,39 +240,36 @@ asyncio.set_event_loop_policy = _new_policy_set
 
 #####
 
-try:
-    _orig_run_get = _aio_event._get_running_loop
+_orig_run_get = _aio_event._get_running_loop
 
-except AttributeError:
-    pass
 
-else:
+def _new_run_get():
+    try:
+        task = trio.lowlevel.current_task()
+    except RuntimeError:
+        pass
+    else:
+        # Trio context. Note: NOT current_loop.get()!
+        # See comment in _TrioPolicy.get_event_loop().
+        return task.context.get(current_loop)
+    # Not Trio context
+    return _orig_run_get()
 
-    def _new_run_get():
-        try:
-            task = trio.lowlevel.current_task()
-        except RuntimeError:
-            pass
-        else:
-            # Trio context. Note: NOT current_loop.get()!
-            # See comment in _TrioPolicy.get_event_loop().
-            return task.context.get(current_loop)
-        # Not Trio context
-        return _orig_run_get()
 
-    # Must override the non-underscore-prefixed get_running_loop() too,
-    # else will use the C-accelerated one which doesn't call the patched
-    # _get_running_loop()
-    def _new_run_get_or_throw():
-        result = _new_run_get()
-        if result is None:
-            raise RuntimeError("no running event loop")
-        return result
+# Must override the non-underscore-prefixed get_running_loop() too,
+# else will use the C-accelerated one which doesn't call the patched
+# _get_running_loop()
+def _new_run_get_or_throw():
+    result = _new_run_get()
+    if result is None:
+        raise RuntimeError("no running event loop")
+    return result
 
-    _aio_event._get_running_loop = _new_run_get
-    _aio_event.get_running_loop = _new_run_get_or_throw
-    asyncio._get_running_loop = _new_run_get
-    asyncio.get_running_loop = _new_run_get_or_throw
+
+_aio_event._get_running_loop = _new_run_get
+_aio_event.get_running_loop = _new_run_get_or_throw
+asyncio._get_running_loop = _new_run_get
+asyncio.get_running_loop = _new_run_get_or_throw
 
 #####
 
@@ -280,6 +299,18 @@ asyncio.get_event_loop = _new_loop_get
 asyncio.set_event_loop = _new_loop_set
 asyncio.new_event_loop = _new_loop_new
 
+# current_task is implemented in C in 3.12+, which creates a problem because it
+# accesses the non-monkeypatched version of _get_running_loop()
+from asyncio import current_task as _orig_current_task
+
+
+def _new_current_task(loop=None):
+    return _orig_current_task(loop or _new_run_get())
+
+
+asyncio.tasks.current_task = _new_current_task
+asyncio.current_task = _new_current_task
+
 #####
 
 
@@ -375,11 +406,33 @@ async def open_loop(queue_len=None):
     """Returns a Trio-flavored async context manager which provides
     an asyncio event loop running on top of Trio.
 
+    The context manager evaluates to a new `TrioEventLoop` object.
+
     Entering the context manager is not enough on its own to immediately
     run asyncio code; it just provides the context that makes running that
     code possible. You additionally need to wrap any asyncio functions
     that you want to run in :func:`aio_as_trio`.
 
+    Exiting the context manager will attempt to do an orderly shutdown
+    of the tasks it contains, analogously to :func:`asyncio.run`.
+    Both asyncio-flavored tasks and Trio-flavored tasks (the latter
+    started using :meth:`~BaseTrioEventLoop.trio_as_future`,
+    :meth:`~BaseTrioEventLoop.run_trio_task`, :func:`trio_as_aio`,
+    etc) are cancelled simultaneously, and the loop waits for them to
+    exit in response to this cancellation before proceeding. All
+    :meth:`~asyncio.loop.call_soon` callbacks that are submitted
+    before exiting the context manager will run before starting this
+    shutdown sequence, and all callbacks that are submitted before the
+    last task exits will run before the loop closes.  The exact point
+    at which the loop stops running callbacks is not specified.
+
+    .. warning:: As with :func:`asyncio.run`, asyncio-flavored tasks
+       that are started *after* exiting the context manager (such as by
+       another task as it unwinds) may or may not be cancelled, and will
+       be abandoned if they survive the shutdown sequence. This may lead
+       to unclosed resources, stderr spew about "coroutine ignored
+       GeneratorExit", etc. Trio-flavored tasks do not have this hazard.
+
     Example usage::
 
             async def async_main(*args):
@@ -403,26 +456,57 @@ async def open_loop(queue_len=None):
         try:
             loop._closed = False
             async with trio.open_nursery() as tasks_nursery:
-                await loop._main_loop_init(tasks_nursery)
-                await loop_nursery.start(loop._main_loop)
-                yield loop
-                tasks_nursery.cancel_scope.cancel()
+                # There are not actually any unshielded checkpoints in
+                # either of the following async functions, so the
+                # shield doesn't do much. However, it is necessary to
+                # make sure that start() actually moves the _main_loop
+                # task into the tasks_nursery if this call to
+                # open_loop() is cancelled. TaskStatus.started()
+                # doesn't complete Nursery.start() if there's a
+                # cancellation pending, because it figures the task
+                # will be cancelled soon enough and doesn't want to
+                # worry about Cancelled exceptions propagating to the
+                # wrong place; but _main_loop shields everything it does
+                # after started(), so this just results in start() never
+                # completing. With the shield here, started() can't see
+                # the outer cancellation, which avoids the deadlock.
+                with trio.CancelScope(shield=True):
+                    await loop._main_loop_init(tasks_nursery)
+                    await loop_nursery.start(loop._main_loop)
 
-                # Allow all submitted run_trio() tasks calls a chance
-                # to start before the tasks_nursery closes, unless the
-                # loop stops (due to someone else calling stop())
-                # before that:
-                async with trio.open_nursery() as sync_nursery:
-                    sync_nursery.cancel_scope.shield = True
+                try:
+                    yield loop
+                finally:
+                    # Allow all already-submitted tasks a chance to start
+                    # (and then immediately be cancelled), unless the loop
+                    # stops (due to someone else calling stop()) before
+                    # that.
+                    async with trio.open_nursery() as sync_nursery:
+                        sync_nursery.cancel_scope.shield = True
 
-                    @sync_nursery.start_soon
-                    async def wait_for_sync():
-                        if not loop.is_closed():
-                            await loop.synchronize()
+                        @sync_nursery.start_soon
+                        async def wait_for_sync():
+                            if not loop.is_closed():
+                                await loop.synchronize()
+                            sync_nursery.cancel_scope.cancel()
+
+                        await loop.wait_stopped()
                         sync_nursery.cancel_scope.cancel()
 
-                    await loop.wait_stopped()
-                    sync_nursery.cancel_scope.cancel()
+                    # Cancel and wait on all currently-running tasks.
+                    # Exiting the tasks_nursery will wait for the Trio tasks
+                    # automatically; we mix in the asyncio tasks by scheduling
+                    # a call to run_aio_future() for each one. It's important
+                    # not to wait on one kind of task before the other, so that
+                    # we support Trio tasks that need to run some asyncio
+                    # code during teardown as well as the opposite.
+                    # Like asyncio.run(), we don't bother cancelling and waiting
+                    # on any additional asyncio tasks that these tasks start
+                    # as they unwind.
+                    aio_tasks = asyncio.all_tasks(loop)
+                    for task in aio_tasks:
+                        tasks_nursery.start_soon(run_aio_future, task)
+                    tasks_nursery.cancel_scope.cancel()
         finally:
             try:
                 await loop._main_loop_exit()
